import pytest

from greg.models import (
    Person,
    Notification,
    Consent,
    Role,
    OrganizationalUnit,
    Sponsor,
    PersonConsent,
    PersonIdentity,
)


@pytest.fixture
def person() -> Person:
    return Person.objects.create(
        first_name="Test",
        last_name="Tester",
        date_of_birth="2000-01-27",
        email="test@example.org",
        mobile_phone="123456789",
    )


@pytest.fixture
def role_foo() -> Role:
    return Role.objects.create(type="role_foo", name_en="Role Foo")


@pytest.fixture
def org_unit_bar() -> OrganizationalUnit:
    return OrganizationalUnit.objects.create(orgreg_id="bar_unit")


@pytest.fixture
def sponsor() -> Sponsor:
    return Sponsor.objects.create(feide_id="sponsor_id")


@pytest.fixture
def consent() -> Consent:
    return Consent.objects.create(
        type="it_guidelines",
        consent_name_en="IT Guidelines",
        consent_name_nb="IT Regelverk",
        consent_description_en="IT Guidelines description",
        consent_description_nb="IT Regelverk beskrivelse",
        consent_link_en="https://example.org/it_guidelines",
        user_allowed_to_change=False,
    )


@pytest.fixture
def person_identity(person: Person) -> PersonIdentity:
    return PersonIdentity.objects.create(
        person=person,
        type=PersonIdentity.IdentityType.PASSPORT,
        source="Test",
        value="12345678901",
    )


@pytest.mark.django_db
def test_person_add_notification(person: Person):
    notifications = Notification.objects.filter(object_type="Person")
    assert len(notifications) == 1
    assert notifications[0].operation == "add"
    assert notifications[0].identifier == person.id


@pytest.mark.django_db
def test_person_update_notification(person: Person):
    person.first_name = "New first name"
    person.save()
    notifications = Notification.objects.filter(object_type="Person")
    assert len(notifications) == 2
    assert notifications[1].operation == "update"
    assert notifications[1].identifier == person.id


@pytest.mark.django_db
def test_person_delete_notification(person: Person):
    person_id = person.id
    person.delete()
    notifications = Notification.objects.filter(object_type="Person")
    assert len(notifications) == 2
    assert notifications[1].operation == "delete"
    assert notifications[1].identifier == person_id


@pytest.mark.django_db
def test_role_add_notification(
    person: Person, role_foo: Role, org_unit_bar: OrganizationalUnit, sponsor: Sponsor
):
    person.roles.add(  # type: ignore
        role_foo,
        through_defaults={
            "start_date": "2021-05-06",
            "end_date": "2021-10-20",
            "unit": org_unit_bar,
            "registered_by": sponsor,
        },
    )
    notifications = Notification.objects.filter(object_type="PersonRole")
    assert len(notifications) == 1
    assert notifications[0].operation == "add"
    meta_data = notifications[0].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["role_id"] == role_foo.id


@pytest.mark.django_db
def test_role_update_notification(
    person: Person, role_foo: Role, org_unit_bar: OrganizationalUnit, sponsor: Sponsor
):
    person.roles.add(  # type: ignore
        role_foo,
        through_defaults={
            "start_date": "2021-05-06",
            "end_date": "2021-10-20",
            "unit": org_unit_bar,
            "registered_by": sponsor,
        },
    )

    assert len(person.person_roles.all()) == 1
    person_role = person.person_roles.all()[0]
    person_role.end_date = "2021-10-21"
    person_role.save()
    notifications = Notification.objects.filter(object_type="PersonRole")
    assert len(notifications) == 2
    assert notifications[1].operation == "update"
    meta_data = notifications[1].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["role_id"] == role_foo.id


@pytest.mark.django_db
def test_role_delete_notification(
    person: Person, role_foo: Role, org_unit_bar: OrganizationalUnit, sponsor: Sponsor
):
    person.roles.add(  # type: ignore
        role_foo,
        through_defaults={
            "start_date": "2021-05-06",
            "end_date": "2021-10-20",
            "unit": org_unit_bar,
            "registered_by": sponsor,
        },
    )

    assert len(person.person_roles.all()) == 1
    person_role = person.person_roles.all()[0]
    person_role.delete()
    notifications = Notification.objects.filter(object_type="PersonRole")
    assert len(notifications) == 2
    assert notifications[1].operation == "delete"
    meta_data = notifications[1].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["role_id"] == role_foo.id


@pytest.mark.django_db
def test_consent_add_notification(person: Person, consent: Consent):
    person.consents.add(consent, through_defaults={"consent_given_at": "2021-06-20"})  # type: ignore
    notifications = Notification.objects.filter(object_type="PersonConsent")
    assert len(notifications) == 1
    assert notifications[0].identifier == person.id
    meta_data = notifications[0].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["consent_id"] == consent.id


@pytest.mark.django_db
def test_consent_update_notification(person: Person, consent: Consent):
    person.consents.add(consent, through_defaults={"consent_given_at": "2021-06-20"})  # type: ignore
    person_consents = PersonConsent.objects.filter(person=person, consent=consent)
    person_consents[0].consent_given_at = "2021-06-21"
    person_consents[0].save()

    notifications = Notification.objects.filter(object_type="PersonConsent")
    assert len(notifications) == 2
    assert notifications[0].identifier == person.id
    meta_data = notifications[0].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["consent_id"] == consent.id


@pytest.mark.django_db
def test_consent_delete_notification(person: Person, consent: Consent):
    person.consents.add(consent, through_defaults={"consent_given_at": "2021-06-20"})  # type: ignore
    person_consents = PersonConsent.objects.filter(person=person, consent=consent)
    person_consents[0].delete()
    notifications = Notification.objects.filter(object_type="PersonConsent")

    assert len(notifications) == 2
    assert notifications[1].identifier == person.id
    assert notifications[1].operation == "delete"
    meta_data = notifications[0].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["consent_id"] == consent.id


@pytest.mark.django_db
def test_person_identity_add_notification(
    person: Person, person_identity: PersonIdentity, sponsor: Sponsor
):
    notifications = Notification.objects.filter(object_type="PersonIdentity")
    assert len(notifications) == 1
    assert notifications[0].identifier == person.id
    assert notifications[0].operation == "add"

    meta_data = notifications[0].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["identity_id"] == person_identity.id


@pytest.mark.django_db
def test_person_identity_update_notification(
    person: Person, person_identity: PersonIdentity, sponsor: Sponsor
):
    person_identity.verified = PersonIdentity.Verified.MANUAL
    person_identity.verified_by = sponsor
    person_identity.verified_when = "2021-08-02"
    person_identity.save()

    notifications = Notification.objects.filter(object_type="PersonIdentity")
    # One notification for adding person identity and one for updating it
    assert len(notifications) == 2
    assert notifications[1].operation == "update"

    meta_data = notifications[1].meta
    assert meta_data["person_id"] == person.id
    assert meta_data["identity_id"] == person_identity.id
